
(* Arbres de syntaxe abstraite issus du parseur *)

type loc = Lexing.position * Lexing.position

type ident = { id_name : string; id_loc : loc }

type p_type =
  | PTboolean
  | PTinteger
  | PTreal
  | PTarray of loc * int * int * p_type
  | PTid of ident

type constant = 
  | Cboolean of bool
  | Cinteger of int
  | Creal of float

type unop = 
  | Unot | Uminus | Uplus 
  | Ureal_of_integer (* seulement au typage *)

type binop = 
  | Beq | Bneq | Blt | Ble | Bgt | Bge 
  | Badd | Bsub | Bmul | Bslash | Bdiv | Bmod
  | Band | Bor

type p_expr =
  { p_expr_node : p_expr_node;
    p_expr_loc : loc }

and p_expr_node =
  | PEconst of constant
  | PElvalue of p_lvalue
  | PEunop of unop * p_expr
  | PEbinop of binop * p_expr * p_expr
  | PEcall of ident * p_expr list

and p_lvalue =
  | PLident of ident
  | PLarrget of p_lvalue * p_expr

type p_pattern = { 
  p_pattern_node : p_pattern_node;
  p_pattern_loc : loc
}

and p_pattern_node = 
  | PPall
  | PPconst of constant
  | PPid of ident
  | PPconstructor of ident * p_pattern list

type p_statement = 
  { p_st_node : p_statement_node;
    p_st_loc : loc }

and p_statement_node =
  | PSskip
  | PSassign of p_lvalue * p_expr
  | PSprocedure_call of ident * p_expr list
  | PSif of p_expr * p_statement * p_statement
  | PSwhile of p_expr * p_statement
  | PSfor of ident * p_expr * p_expr * p_statement
  | PSbloc of p_statement list
  | PSmatch of p_expr * (p_pattern * p_statement list) list

type p_decl_vars = (ident * p_type) list

type p_constructor = ident * p_type list

type p_decl_types = (ident * p_constructor list) list

type mode = Value | Reference

type p_parameter = mode * ident * p_type

type p_definition_type = 
  | PDfunction of p_type
  | PDprocedure

type p_definition = 
  ident * p_parameter list * p_definition_type * p_decl_vars * p_statement

type p_file = { 
  p_name : ident;
  p_decl_types : p_decl_types;
  p_decl_vars : p_decl_vars;
  p_definitions : p_definition list;
  p_main : p_statement;
}

(* Arbres de syntaxe abstraite typs *)

type t_type =
  | Tboolean
  | Tinteger
  | Treal
  | Tarray of int * int * t_type
  | Tid of string

type a_constructor = {
  c_name : string;
  c_args : t_type list;
  c_type : string;
  c_id : int; (* unique *)
}

type a_type = { 
  a_name : string;
  a_constructors : a_constructor list
}

type global_var = {
  gv_name : string;
  gv_type : t_type;
  gv_id : int; (* unique *)
}

type local_var = {
  lv_name : string;
  lv_type : t_type;
  lv_mode : mode;
  lv_id : int; (* unique *)
}

type t_function = {
  f_name : string;
  f_params : local_var list;
  f_type : t_type;
}

type t_procedure = {
  pr_name : string;
  pr_params : local_var list;
}

type constructor = 
  | Cconst of constant
  | Cnamed of a_constructor

type t_pattern = 
  | Pall
  | Pid of local_var
  | Pconstructor of constructor * t_pattern list

type t_expr =
  { t_expr_node : t_expr_node;
    t_expr_type : t_type }

and t_expr_node =
  | Econst of constant
  | Elvalue of t_lvalue
  | Eunop of unop * t_expr
  | Ebinop of binop * t_expr * t_expr
  | Ecall of t_function * (mode * t_expr) list
  | Econstr of a_constructor * t_expr list

and t_lvalue =
  | Llocal of local_var
  | Lglobal of global_var
  | Larrget of t_lvalue * t_expr * array_info

and array_info = int (* premier indice *) * t_type (* type des lments *)

(* Type pour le pseudo-code intermdiaire construit par Matching *)

type value = (* les valeurs filtres *)
  | Top
  | Field of int * value

type 'a action =
  | User of 'a
  | Set of local_var * value * 'a action

type 'a code = 
  | Action of 'a action
  | Case of value * (constructor * 'a code) list * 'a code option

type t_statement = 
  | Sskip
  | Sassign of t_lvalue * t_expr
  | Sassign_result of t_expr
  | Sprocedure_call of t_procedure * (mode * t_expr) list
  | Sif of t_expr * t_statement * t_statement
  | Swhile of t_expr * t_statement
  | Sfor of t_lvalue * local_var * t_expr * t_expr * t_statement
  | Sbloc of t_statement list
  (*| Smatch of t_expr * (t_pattern * t_statement list) list*)
  | Smatch of t_expr * (t_statement list) code
  | Swriteln of t_expr list

type body = {
  b_vars : local_var list;
  b_stmt : t_statement
}

type t_definition = 
  | Dfunction of t_function * body
  | Dprocedure of t_procedure * body

type t_file = {
  t_name : string;
  t_decl_types : a_type list;
  t_decl_vars : global_var list;
  t_definitions : t_definition list;
  t_main : t_statement;
}

