
(* Production de code *)

open Ast
open Typing
open Instr
open Emit
open Format

let debug = ref false

(* Primitives *)

let emit_primitive = function
  | "trunc" -> emit [Ftoi]
  | "cos" -> emit [Fcos]
  | "sin" -> emit [Fsin]
  | "open_screen" -> emit [OpenDrawingArea]
  | "clear_screen" -> emit [ClearDrawingArea]
  | "set_color" -> emit [SetColor]
  | "draw_point" -> emit [DrawPoint]
  | "draw_line"-> emit [DrawLine]
  | "draw_circle" -> emit [DrawCircle]
  | "draw_rect" -> emit [DrawRect]
  | "fill_rect" -> emit [FillRect]
  | "usleep" -> emit [Refresh; Pusha "__usleep"; Call; Pop 1]
  | _ -> assert false

(* Gnration d'tiquettes fraches *)

let new_label = 
  let r = ref 0 in fun () -> incr r; "__label_" ^ string_of_int !r

(* Taille des donnes sur la pile 
   Renvoie aussi le type de base pour l'initialisation 
   (il faut savoir s'il faut initialiser avec 0.0 ou 0) *)

let rec tsize = function
  | Tboolean | Tinteger | Tid _ -> 1, Tinteger
  | Treal -> 1, Treal
  | Tarray (n1, n2, ty) -> let s,t = tsize ty in (n2 - n1 + 1) * s, t

let size ty = fst (tsize ty)

let size_param x = match x.lv_mode with
  | Value -> size x.lv_type
  | Reference -> 1

let size_params = List.fold_left (fun n x -> n + size_param x) 0

(* Tables indexes par des entiers uniques *)

module type Unique = sig type t val id : t -> int end

module Hunique(U : Unique) = struct 
  module X = struct
    type t = U.t
    let hash x = abs (U.id x)
    let equal x y = U.id x = U.id y
  end
  include Hashtbl.Make(X)
  let t = create 97
  let add = add t
  let find = find t
end

module Local = Hunique(struct type t = local_var let id v = v.lv_id end)

module Global = Hunique(struct type t = global_var let id v = v.gv_id end)

(* Oprateurs *)

let emit_comp b ty = match b, ty with
  | Beq, _ -> emit [Equal]
  | Bneq, _ -> emit [Equal; Not]
  | Blt, (Tboolean | Tinteger) -> emit [Inf]     
  | Blt, Treal -> emit [Finf]
  | Ble, (Tboolean | Tinteger) -> emit [Infeq]
  | Ble, Treal -> emit [Finfeq]
  | Bgt, (Tboolean | Tinteger) -> emit [Sup]
  | Bgt, Treal -> emit [Fsup]
  | Bge, (Tboolean | Tinteger) -> emit [Supeq]
  | Bge, Treal -> emit [Fsupeq]
  | _ -> assert false

let emit_arith b ty = match b, ty with
  | Badd, Tinteger -> emit [Add]
  | Badd, Treal -> emit [Fadd]
  | Bsub, Tinteger -> emit [Sub]
  | Bsub, Treal -> emit [Fsub]
  | Bmul, Tinteger -> emit [Mul]
  | Bmul, Treal -> emit [Fmul]
  | Bdiv, Tinteger -> emit [Div]
  | Bslash, Treal -> emit [Fdiv]
  | Bmod, Tinteger -> emit [Mod]
  | _ -> assert false

(* est-ce qu'un constructeur est allou sur le tas ? *)

let on_heap = function
  | Cconst _ -> false
  | Cnamed { c_args = [] } -> false
  | Cnamed _ -> true

(* Expressions. 
   [compile e] compile la valeur de l'expression [e] en sommet de pile. *)

let compile_constant = function
  | Cboolean true -> Pushi 1
  | Cboolean false -> Pushi 0
  | Cinteger i -> Pushi i
  | Creal f -> Pushf f

let rec compile_expr e = match e.t_expr_node with
  | Econst c ->
      emit [compile_constant c]
  (* optim *)
  | Elvalue (Llocal lv) ->
      emit [Pushl (Local.find lv)];
      if lv.lv_mode = Reference then emit [Load 0]
  (* optim *)
  | Elvalue (Lglobal gv) ->
      emit [Pushg (Global.find gv)]
  | Elvalue lv ->
      compile_lvalue lv;
      emit [Load 0]
  | Eunop (Uplus, e) ->
      compile_expr e
  | Eunop (Uminus, e) ->
      begin match e.t_expr_type with
	| Tinteger -> emit [Pushi 0]; compile_expr e; emit [Sub]
	| Treal -> emit [Pushf 0.0]; compile_expr e; emit [Fsub]
	| _ -> assert false
      end
  | Eunop (Unot, e) ->
      compile_expr e;
      emit [Not]
  | Eunop (Ureal_of_integer, e) ->
      compile_expr e;
      emit [Itof]
  | Ebinop (Band, e1, e2) ->
      compile_expr e1;
      let l = new_label () in
      emit [Dup 1; Jz l; Pop 1];
      compile_expr e2;
      emit [Dup 1; Jz l; Pop 1; Pushi 1];
      emit [Label l]
  | Ebinop (Bor, e1, e2) ->
      compile_expr e1;
      let l = new_label () in
      emit [Pushi 1; Swap; Not; Jz l; Pop 1];
      compile_expr e2;
      emit [Dup 1; Jz l; Pop 1; Pushi 1];
      emit [Label l]
  | Ebinop (Beq | Bneq as op, e1, e2) ->
      assert (e1.t_expr_type = e1.t_expr_type);
      begin match e1.t_expr_type with
	| Tid _ -> 
	    emit [Pushn 1]; compile_expr e1; compile_expr e2; 
	    emit [Pusha "__compare"; Call; Pop 2]
	| ty1 -> 
	    compile_expr e1; compile_expr e2; emit [Equal]
      end;
      if op = Bneq then emit [Not]
  | Ebinop (Blt | Ble | Bgt | Bge as op, e1, e2) ->
      assert (e1.t_expr_type = e1.t_expr_type);
      compile_expr e1; 
      compile_expr e2; 
      emit_comp op e1.t_expr_type
  | Ebinop (Badd | Bsub | Bmul | Bdiv | Bslash | Bmod as op, e1, e2) ->
      assert (e1.t_expr_type = e1.t_expr_type);
      compile_expr e1;
      compile_expr e2;
      emit_arith op e1.t_expr_type
  | Ecall (f, [_,e]) when is_primitive_function f ->
      compile_expr e;
      emit_primitive f.f_name
  | Ecall (f, el) ->
      emit [Pushn 1];
      List.iter compile_arg el;
      emit [Pusha ("_" ^ f.f_name); Call];
      let n = size_params f.f_params in
      if n > 0 then emit [Pop n]
  | Econstr (c, []) ->
      emit ~comment:("constr. " ^ c.c_name) [Pushi c.c_id]
  | Econstr (c, el) ->
      let n = List.length el in
      emit [Alloc (n+1); Dup 1];
      emit ~comment:("constr. " ^ c.c_name) [Pushi c.c_id]; emit [Store 0];
      let i = ref 0 in
      List.iter 
	(fun e -> incr i; emit [Dup 1]; compile_expr e; emit [Store !i]) el

and compile_lvalue = function
  | Llocal x ->
      let i = Local.find x in
      if x.lv_mode = Reference then 
	emit [Pushl i] 
      else 
	emit [Pushfp; Pushi i; Padd]
  | Lglobal x ->
      emit [Pushgp; Pushi (Global.find x); Padd]
  | Larrget (lv, e, (n1, ty)) ->
      compile_lvalue lv;
      compile_expr e;
      emit [Pushi n1; Sub; Pushi (size ty); Mul; Padd]

and compile_arg (m, e) = match m, e.t_expr_type, e.t_expr_node with
  | Value, (Tarray _ as ty), Elvalue lv -> 
      let s = size ty in
      emit [Pushn s; Pushsp; Pushi (-s); Padd];
      compile_lvalue lv;
      emit [Pushi s; Pusha "__memcopy"; Call; Pop 3]
  | Value, _, _ -> 
      compile_expr e
  | Reference, _, Elvalue lv -> 
      compile_lvalue lv
  | _ -> 
      assert false
      
      
(* Instructions *)

let rec compile_statement ret = function
  | Sskip ->
      ()
  (* optim *)
  | Sassign (Llocal lv, e) when lv.lv_mode = Value -> 
      compile_expr e;
      emit [Storel (Local.find lv)]
  (* optim *)
  | Sassign (Lglobal gv, e) -> 
      compile_expr e;
      emit [Storeg (Global.find gv)]
  | Sassign (lv, e) -> 
      compile_lvalue lv;
      compile_expr e;
      emit [Store 0]
  | Sassign_result e -> 
      compile_expr e;
      emit [Storel ret]
  | Sprocedure_call (p, el) when is_primitive_procedure p ->
      List.iter compile_arg el;
      emit_primitive p.pr_name
  | Sprocedure_call (p, el) ->
      List.iter compile_arg el;
      emit [Pusha ("_" ^ p.pr_name); Call];
      let n = size_params p.pr_params in
      if n > 0 then emit [Pop n]
  | Sif (e, s1, s2) ->
      compile_expr e;
      let l1 = new_label () in
      let l2 = new_label () in
      emit [Jz l1];
      compile_statement ret s1;
      emit [Jump l2; Label l1];
      compile_statement ret s2;
      emit [Label l2]
  | Swhile (e, s) ->
      let l1 = new_label () in
      let l2 = new_label () in
      emit [Label l1];
      compile_expr e;
      emit [Jz l2];
      compile_statement ret s;
      emit [Jump l1; Label l2]
  | Sfor (x, v2, e1, e2, s) -> 
      (* sucre syntaxique pour
         x := e1; v2 := e2; while i <= v2 do begin s; i := i + 1 *)
      let ex = { t_expr_node = Elvalue x; t_expr_type = Tinteger } in
      let lv2 = Llocal v2 in
      let ev2 = { t_expr_node = Elvalue lv2; t_expr_type=Tinteger } in
      let one = { t_expr_node = Econst (Cinteger 1); t_expr_type=Tinteger } in
      let s = 
	Sbloc 
	  [Sassign (x, e1);
	   Sassign (lv2, e2);
	   Swhile ({ t_expr_node=Ebinop (Ble, ex, ev2); t_expr_type=Tboolean },
		   Sbloc
		     [s; Sassign (x,
				  { t_expr_node = Ebinop (Badd, ex, one);
				    t_expr_type = Tinteger })])]
      in
      compile_statement ret s
  | Sbloc sl ->
      List.iter (compile_statement ret) sl
  | Smatch (e, bl) ->
      if !debug then eprintf "%a@." Matching.print bl;
      let rec value = function
	| Top -> emit [Dup 1]
	| Field (i, v) -> value v; emit [Load i]
      in
      let rec action = function
	| User sl -> emit [Pop 1]; List.iter (compile_statement ret) sl
	| Set (x, v, a) -> value v; emit [Storel (Local.find x)]; action a
      in
      let rec code = function
	| Action a -> 
	    action a
	| Case (v, bl, other) ->
	    let end_case = new_label () in
	    value v;
	    List.iter
	      (fun (ci, ccode) ->
		 let next = new_label () in
		 let push_tag = function
		   | Cconst c -> compile_constant c
		   | Cnamed c -> Pushi c.c_id
		 in
		 emit [Dup 1]; 
		 if on_heap ci then emit [Dup 1; Is_addr; Jz next; Load 0];
		 emit [push_tag ci; Equal; Jz next; Pop 1];
		 code ccode;
		 emit [Jump end_case; Label next]) 
	      bl;
	    emit [Pop 1];
	    begin match other with
	      | None -> ()
	      | Some c -> code c
	    end;
	    emit [Label end_case]
      in
      compile_expr e;
      code bl
  | Swriteln el -> 
      let compile_one e = 
	compile_expr e;
	match e.t_expr_type with
	  | Tboolean -> emit [Pusha "__print_boolean"; Call]
	  | Tinteger -> emit [Writei]
	  | Treal -> emit [Writef]
	  | _ -> assert false
      in
      List.iter compile_one el

(* Variables globales (allocation) *)

let initialize s = function
  | Tinteger -> emit [Pushn s]
  | Treal when s = 1 -> emit [Pushf 0.0]
  | Treal -> emit [Pushn s; Pushi s; Pusha "__fill_zero_float"; Call; Pop 1]
  | _ -> assert false

let gp = ref 0

let global_var gv = 
  emit [Comment ("global var " ^ gv.gv_name)];
  let s,t = tsize gv.gv_type in
  initialize s t;
  if !debug then eprintf "global %s en GP[%d]@." gv.gv_name !gp;
  Global.add gv !gp;
  gp := !gp + s

(* Fonctions et procdures *)

let alloc_locals b =
  let n = ref 0 in
  let alloc_local i x = 
    if !debug then eprintf "local %s en FP[%d]@." x.lv_name i;
    Local.add x i; 
    let s,t = tsize x.lv_type in
    let i' = i + s in
    if i' >= !n then begin initialize (i' - !n) t; n := i' end;
    i'
  in
  let rec action i = function
    | User sl -> bloc i sl
    | Set (x, _, a) -> let i' = alloc_local i x in action i' a
  and code i = function
    | Action a -> action i a
    | Case (_, bl, None) -> List.iter (fun (_,c) -> code i c) bl
    | Case (_, bl, Some c) -> List.iter (fun (_,c) -> code i c) bl; code i c
  and match_vars i = function
    | Smatch (_, c) -> code i c
    | Sif (_, s1, s2) -> match_vars i s1; match_vars i s2
    | Sfor (_, v2, _, _, s) -> let i' = alloc_local i v2 in match_vars i' s
    | Swhile (_, s) -> match_vars i s
    | Sbloc sl -> bloc i sl
    | Swriteln _ | Sprocedure_call _ | Sassign_result _ | Sassign _ | Sskip ->
	()
  and bloc i sl = 
    List.iter (match_vars i) sl
  in
  let i = List.fold_left alloc_local 0 b.b_vars in
  match_vars i b.b_stmt

let alloc_params pl =
  List.fold_right 
    (fun x i -> 
       let s = size_param x in
       if !debug then eprintf "param %s en FP[%d]@." x.lv_name (i-s);
       Local.add x (i-s);
       i-s) 
    pl 0

let definition = function
  | Dfunction (f, b) ->
      emit [Label ("_" ^ f.f_name)];
      alloc_locals b;
      let n = alloc_params f.f_params in
      compile_statement (n - 1) b.b_stmt;
      emit [Return]
  | Dprocedure (p, b) ->
      emit [Label ("_" ^ p.pr_name)];
      alloc_locals b;
      ignore (alloc_params p.pr_params);
      compile_statement 0 b.b_stmt;
      emit [Return]

let compile file p = 
  emit [Comment ("program " ^ p.t_name)];
  List.iter global_var p.t_decl_vars;
  emit [Start];
  emit [Jump "_main"];
  List.iter definition p.t_definitions;
  emit [Label "_main"];
  alloc_locals { b_vars = []; b_stmt = p.t_main };
  compile_statement 0 p.t_main;
  emit [Stop];
  (* memcopy : attend n, src et dst en sommet de pile
     et copie n mots  partir de src en dst *)
  let debut = new_label () in
  let fin = new_label () in
  emit [Label "__memcopy"; 
	Label debut; Dup 1; Jz fin;
	Pushi 1; Sub; Dup 2; Loadn; (* src[n]:n:src:dst *)
	Pushsp; Load (-4); Pushsp; Load (-3); Padd; (* dst+n:src[n]:n... *)
	Swap; Instr.Store 0; Jump debut;
	Label fin; Return];
  emit [Label "__usleep";
	Dup 1; Jz "__usleep_end";
	Pushi 1; Sub; Jump "__usleep";
	Label "__usleep_end"; Return];
  emit [Label "__print_boolean";
	Jz "__print_false";
	Pushs "true"; Writes; Return;
	Label "__print_false";
	Pushs "false"; Writes; Return];
  (* compare : prend deux valeurs d'un mme type et les compare *)
  let v1_prim = new_label () in
  let false_ = new_label () in
  let true_ = new_label () in
  let boucle = new_label () in
  let deux_prim = new_label () in
  emit [Label "__compare";
	Pushl (-2); Is_addr; Jz v1_prim;
	Pushl (-1); Is_addr; Jz false_; (* v1 addr et v2 prim *)
	(* deux addr. *)
	Pushl (-2); Load 0; (* tag v1 *) Dup 1;
	Pushl (-1); Load 0; (* tag v2 *) Equal; Jz false_; (* tags diff *)
	(* mme tags *)
	Pushi tag_multiplier; Div; (* taille n en sommet de pile *)
	Label boucle; Dup 1; Jz true_;
	Pushn 1;
	Pushl (-2); Pushl 0; Loadn;
	Pushl (-1); Pushl 0; Loadn;
	Pusha "__compare"; Call; Pop 2;
	Jz false_;
	Pushl 0; Pushi 1; Sub; Storel 0; Jump boucle;
	Label true_; Pushi 1; Storel (-3); Return;
	Label v1_prim; Pushl (-1); Is_addr; Jz deux_prim;
	(* v1 prim et v2 addr *)
	Label false_; Return;
	Label deux_prim; Pushl (-1); Pushl (-2); Equal; Storel (-3);
	Return];
  (* fill_zero_float : 
     attend un argument n en FP[-1] et met  0.0 FP[-2..-n-1] *)
  let fin = new_label () in
  emit [Label "__fill_zero_float";
	Pushl (-1); Jz fin; (* n=0 ? *)
	Pushfp; Pushl (-1); Dup 1; Pushi 1; Sub; Storel (-1); (* n=n-1 *)
	Pushi (-1); Swap; Sub; (* -n-1 *)
	Pushf 0.0; Storen; (* FP[n-1]=0.0 *)
	Jump "__fill_zero_float";
	Label fin; Return];
  let vm = (Filename.chop_suffix file ".pas" ^ ".vm") in
  write_file vm


