
open Ast

let types = Hashtbl.create 17

let add_type ty = Hashtbl.add types ty.a_name ty

exception NotExhaustive

module Cmap = Map.Make(struct type t = constructor let compare = compare end)

let args = function Cconst _ -> [] | Cnamed c -> c.c_args

let is_var = function Pall | Pid _ -> true | _ -> false

(* indique si [n] est le nb de constructeurs de [ty] *)
let not_enough_for n = function
  | Tboolean -> n < 2
  | Tinteger | Treal -> true (* lger abus *)
  | Tarray _ -> assert false
  | Tid x ->
      let d = Hashtbl.find types x in
      n < List.length d.a_constructors

(* Compilation du filtrage *)

type 'a row = {
  elements : (t_type * t_pattern) list;
  action : 'a action
}

type 'a matrix = {
  values : value list;
  rows : 'a row list;
}

(* projection [t1;...;tn] v = [ Field(1,v); ...; Field(n,v) ] *)
let projections tl v =
  let i = ref 0 in
  List.map (fun _ -> incr i; Field (!i, v)) tl

let update_action v1 r = match r.elements with
  | (_, Pid x) :: _ -> Set (x, v1, r.action)
  | (_, Pall) :: _ -> r.action
  | _ -> assert false 

let compile ty pl =
  (* la matrice est reprsente par la liste de ses lignes *)
  let rec matrix m = match m.rows with
    | [] -> 
	(* aucune ligne *)
	raise NotExhaustive
    | { elements = []; action = a } :: _ -> 
	(* au moins une action : OK *)
	Action a
	(* note : s'il y en a plus, cela ne signifie pas ncessairement
	   que la clause est inutile car l'algo duplique des actions *)
    | { elements = (ty,_) :: _ } :: _ ->
	let v1 = List.hd m.values in
	if List.for_all (fun r -> is_var (snd (List.hd r.elements))) m.rows 
	then
	  let update_row r =
	    { elements = List.tl r.elements; action = update_action v1 r }
	  in
	  matrix { values = List.tl m.values;
		   rows = List.map update_row m.rows }
	else
	  (* on spare les lignes par constructeurs *)
	  let empty c =
	    { values = projections (args c) v1 @ List.tl m.values;
	      rows = [] }
	  in
	  let cm0 = 
	    List.fold_left 
	      (fun cm r -> match r.elements with
		 | (_, Pconstructor (c, _)) :: _ -> Cmap.add c (empty c) cm
		 | (_, (Pall | Pid _)) :: _ -> cm
		 | [] -> assert false)
	      Cmap.empty m.rows
	  in
	  let dispatch r cm = match r.elements with
	    | [] -> 
		assert false
	    | (_, (Pall | Pid _ as p)) :: rr ->
		(* une variable => on doit la reprendre pour chaque constr. *)
		Cmap.fold
		  (fun c m cm -> 
		     let h = List.map (fun ty -> ty, Pall) (args c) in
		     let r' = 
		       { elements = h @ rr; action = update_action v1 r } 
		     in
		     Cmap.add c { m with rows = r' :: m.rows } cm)
		  cm cm
	    | (_, (Pconstructor (c, pl))) :: rr ->
		let h = List.combine (args c) pl in
		let m = Cmap.find c cm in
		let r' = { elements = h @ rr; action = r.action } in
		Cmap.add c { m with rows = r' :: m.rows} cm
	  in
	  let cm = List.fold_right dispatch m.rows cm0 in
	  let nbc = ref 0 in
	  let cases = 
	    Cmap.fold (fun c m acc -> incr nbc; (c, matrix m) :: acc) cm []
	  in
	  let otherwise = 
	    if not_enough_for !nbc ty then
	      let variable r mr = match r.elements with
		| [] -> 
		    assert false
		| (_, (Pall | Pid _)) :: rr -> 
		    let r' = { elements = rr; action = update_action v1 r } in
		    r' :: mr
		| (_, Pconstructor _) :: _ -> 
		    mr
	      in
	      let mr = { values = List.tl m.values;
			 rows = List.fold_right variable m.rows [] } in
	      Some (matrix mr)
	    else 
	      None
	  in
	  Case (v1, cases, otherwise)
  in
  matrix 
    { values = [Top];
      rows = List.map (fun (p,s) -> {elements = [ty, p]; action = User s}) pl }

(* debug *)

open Format

let constructor fmt = function
  | Cconst (Cboolean b) -> fprintf fmt "%b" b
  | Cconst (Cinteger n) -> fprintf fmt "%d" n
  | Cconst (Creal f) -> fprintf fmt "%f" f
  | Cnamed c -> fprintf fmt "%s" c.c_name

let print fmt c = 
  let rec value fmt = function
    | Top -> fprintf fmt "top"
    | Field (i, v) -> fprintf fmt "#%d(%a)" i value v
  in
  let rec action fmt = function
    | User _ -> fprintf fmt "<user>"
    | Set (x, v, a) -> fprintf fmt "set %s = %a; %a" x.lv_name value v action a
  in
  let rec code fmt = function
    | Action a ->
	action fmt a
    | Case (v, bl, other) ->
	fprintf fmt "@[case %a of@\n" value v;
	List.iter (fun (cstr, c) -> 
		     fprintf fmt "%a: %a@\n" constructor cstr code c) bl;
	begin match other with 
	  | None -> ()
	  | Some c -> fprintf fmt "other: %a@\n" code c
	end;
	fprintf fmt "end@]"
  in
  fprintf fmt "@[%a@]" code c

  
