
(* Analyse smantique *)

open Ast
open Matching
open Format

type error =
  | ExpectedType of t_type * t_type
  | ExpectedPattern of t_type
  | ExpectedNum of t_type
  | UnboundType of string
  | UnboundVar of string
  | UnboundFun of string
  | UnboundProc of string
  | UnboundConstr of string
  | Shadow of string
  | TooFewArguments
  | Clash of string
  | NoReturn
  | Other of string

exception Error of loc * error
let dummy_loc = Lexing.dummy_pos, Lexing.dummy_pos

let error loc e = raise (Error (loc, e))
let errors loc s = error loc (Other s)

let rec print_type fmt = function
  | Tboolean -> fprintf fmt "boolean"
  | Tinteger -> fprintf fmt "integer"
  | Treal -> fprintf fmt "real"
  | Tarray (f, t, ty) -> fprintf fmt "array [%d..%d] of %a" f t print_type ty
  | Tid s -> fprintf fmt "%s" s

let report fmt = function
  | UnboundType id -> fprintf fmt "unbound type %s" id
  | UnboundVar id -> fprintf fmt "unbound variable %s" id
  | UnboundFun id -> fprintf fmt "unbound function or constructor %s" id
  | UnboundProc id -> fprintf fmt "unbound procedure %s" id
  | UnboundConstr id -> fprintf fmt "unbound constructor %s" id
  | ExpectedType (t1,t2) -> 
      fprintf fmt 
      "this expression has type %a but is expected to have type %a" 
      print_type t1 print_type t2
  | ExpectedPattern ty ->
      fprintf fmt "this pattern is expected to have type %a" 
	print_type ty
  | ExpectedNum ty -> 
      fprintf fmt 
     "this expression has type %a but is expected to have type integer or real"
      print_type ty
  | Clash id -> fprintf fmt "redeclaration of %s" id
  | Shadow id -> fprintf fmt "shadowing function-identifier %s" id
  | TooFewArguments -> fprintf fmt "too few arguments"
  | NoReturn -> fprintf fmt "missing returned value in this function"
  | Other s -> fprintf fmt "%s" s

(* [fold_map f l acc] calcule  la fois [map f l] avec la premire projection 
   du rsultat de [f] et [fold f l acc] avec la seconde projection de [f] *)
let fold_map f l acc =
  let rec fold fl acc = function
    | [] -> List.rev fl, acc
    | x :: l -> let fx,acc = f x acc in fold (fx :: fl) acc l
  in
  fold [] acc l

(* Entiers uniques pour la table des symboles *)

let unique_id = let r = ref (-1) in fun () -> incr r; !r

(* Tables des symboles *)

let types = Hashtbl.create 17
let constructors = Hashtbl.create 97
let global_vars = Hashtbl.create 97
let functions = Hashtbl.create 97
let procedures = Hashtbl.create 97

let is_global id = 
  Hashtbl.mem types id ||
  Hashtbl.mem constructors id ||
  Hashtbl.mem global_vars id ||
  Hashtbl.mem functions id ||
  Hashtbl.mem procedures id

(* Types *)

let rec type_type = function
  | PTboolean ->
      Tboolean
  | PTinteger ->
      Tinteger
  | PTreal ->
      Treal
  | PTarray (loc, i1, i2, ty) ->
      if i1 > i2 then errors loc "invalid subrange type";
      Tarray (i1, i2, type_type ty)
  | PTid { id_name = id; id_loc = loc } ->
      if not (Hashtbl.mem types id) then error loc (UnboundType id);
      Tid id

let is_array_type = function Tarray _ -> true | _ -> false

let not_id = function Tid _ -> false | _ -> true

let tag_multiplier = 1000000

let constructor ty ({id_name=id; id_loc=loc}, al) =
  if is_global id then error loc (Clash id);
  let type_arg ty =
    let ty = type_type ty in
    if is_array_type ty then 
      errors loc "array not allowed as constructor component";
    ty
  in
  let tag = (List.length al) * tag_multiplier + unique_id () in
  let c = 
    { c_name = id; c_args = List.map type_arg al; 
      c_type = ty; c_id = tag }
  in
  Hashtbl.add constructors id c;
  c
  
let decl_type ({id_name=id; id_loc=loc}, cl) =
  (* on peut se contenter de regarder dans la seule table [types]
     car les types viennent forcment avant les autres globaux *)
  if Hashtbl.mem types id then error loc (Clash id);
  Hashtbl.add types id ();
  let ty = { a_name = id; a_constructors = List.map (constructor id) cl } in
  Matching.add_type ty;
  ty

(* Compatibilits entre types et coercions entiers vers flotants *)

let sup_type t1 t2 = match t1, t2 with
  | Tboolean, Tboolean ->
      Tboolean
  | Tinteger, Tinteger -> 
      Tinteger
  | Tinteger, Treal
  | Treal, Tinteger
  | Treal, Treal -> 
      Treal
  | Tid n1, Tid n2 ->
      assert (n1 = n2);
      t1
  | _ ->
      assert false

let compatible t1 t2 = match t1, t2 with
  | Tboolean, Tboolean
  | Tinteger, Tinteger 
  | Tinteger, Treal
  | Treal, Tinteger
  | Treal, Treal -> 
      true
  | Tid n1, Tid n2 ->
      n1 = n2
  | _ ->
      false

let coerce loc e ty =
  if e.t_expr_type = ty then 
    e 
  else if e.t_expr_type = Tinteger && ty = Treal then
    { t_expr_node = Eunop (Ureal_of_integer, e); t_expr_type = Treal }
  else
    error loc (ExpectedType (e.t_expr_type, ty))

(* Fonctions et procdures prdfinies *)

let value_var x ty = 
  { lv_name = x; lv_type = ty; lv_mode = Value; lv_id = unique_id () }

let x_real = value_var "x" Treal
let x_integer = value_var "x" Tinteger
let y_integer = value_var "y" Tinteger
let z_integer = value_var "z" Tinteger
let t_integer = value_var "t" Tinteger

let trunc = { f_name = "trunc"; f_params = [x_real]; f_type = Tinteger }
let cos = { f_name = "cos"; f_params = [x_real]; f_type = Treal }
let sin = { f_name = "sin"; f_params = [x_real]; f_type = Treal }

let primitive_functions = [trunc; cos; sin]

let () = 
  List.iter (fun f -> Hashtbl.add functions f.f_name f) primitive_functions

let is_primitive_function f = List.memq f primitive_functions

let open_screen = 
  { pr_name = "open_screen"; pr_params = [x_integer; y_integer] }
let clear_screen = 
  { pr_name = "clear_screen"; pr_params = [] }
let set_color = 
  { pr_name = "set_color"; 
    pr_params = [x_integer; y_integer; z_integer] }
let draw_point = 
  { pr_name = "draw_point"; pr_params = [x_integer; y_integer] }
let draw_line = 
  { pr_name = "draw_line"; 
    pr_params = [x_integer; y_integer; z_integer; t_integer] }
let draw_circle = 
  { pr_name = "draw_circle"; 
    pr_params = [x_integer; y_integer; z_integer] }
let draw_rect = 
  { pr_name = "draw_rect"; 
    pr_params = [x_integer; y_integer; z_integer; t_integer] }
let fill_rect = 
  { pr_name = "fill_rect"; 
    pr_params = [x_integer; y_integer; z_integer; t_integer] }
let usleep = 
  { pr_name = "usleep"; pr_params = [x_integer] }

let primitive_procedures =
  [ open_screen; clear_screen; set_color;
    draw_point; draw_line; draw_circle; draw_rect; fill_rect; usleep ]

let () = 
  List.iter (fun p -> Hashtbl.add procedures p.pr_name p) primitive_procedures

let is_primitive_procedure p = List.memq p primitive_procedures

(* Expressions *)

module Env = Map.Make(String)

let type_constant = function
  | Cboolean _ -> Tboolean
  | Cinteger _ -> Tinteger
  | Creal _    -> Treal

let rec type_expr env e = 
  let e,t = type_expr_node env e.p_expr_loc e.p_expr_node in
  { t_expr_node = e; t_expr_type = t }

and type_expr_node env loc = function
  | PEconst c ->
      Econst c, type_constant c
  | PElvalue (PLident {id_name=x} as lv) ->
      begin 
	try 
	  let lv, ty = type_lvalue env loc lv in Elvalue lv, ty
	with Error _ -> try
	  let f = Hashtbl.find functions x in
	  if f.f_params <> [] then error loc (TooFewArguments);
	  Ecall (f, []), f.f_type
	with Not_found -> try
	  let c = Hashtbl.find constructors x in
	  if c.c_args <> [] then error loc (TooFewArguments);
	  Econstr (c, []), Tid c.c_type
	with Not_found ->
	  error loc (UnboundVar x)
      end
  | PElvalue lv -> 
      let lv, ty = type_lvalue env loc lv in
      Elvalue lv, ty
  | PEunop (Unot, e) ->
      let e = type_boolean_expr env e in
      Eunop (Unot, e), Tboolean
  | PEunop ((Uplus | Uminus as op), e) ->
      let e = type_num_expr env e in
      Eunop (op, e), e.t_expr_type
  | PEunop (Ureal_of_integer, _) ->
      assert false
  | PEbinop ((Band | Bor as op), e1, e2) ->
      let e1 = type_boolean_expr env e1 in
      let e2 = type_boolean_expr env e2 in
      Ebinop (op, e1, e2), Tboolean
  | PEbinop ((Badd | Bsub | Bmul as op), e1, e2) ->
      let e1 = type_num_expr env e1 in
      let e2 = type_num_expr env e2 in
      let ty = sup_type e1.t_expr_type e2.t_expr_type in
      Ebinop (op, coerce loc e1 ty, coerce loc e2 ty), ty
  | PEbinop (Bslash, e1, e2) ->
      let e1 = type_num_expr env e1 in
      let e2 = type_num_expr env e2 in
      Ebinop (Bslash, coerce loc e1 Treal, coerce loc e2 Treal), Treal
  | PEbinop ((Bdiv | Bmod as op), e1, e2) ->
      let e1 = type_int_expr env e1 in
      let e2 = type_int_expr env e2 in
      Ebinop (op, e1, e2), Tinteger
  | PEbinop (Beq | Bneq | Blt | Ble | Bgt | Bge as op, e1, e2) ->
      let e1 = type_expr env e1 in
      let ty1 = e1.t_expr_type in
      let e2 = type_expr env e2 in
      let ty2 = e2.t_expr_type in
      if compatible ty1 ty2 && (op = Beq || op = Bneq || not_id ty1) then
	let ty = sup_type ty1 ty2 in
	Ebinop (op, coerce loc e1 ty, coerce loc e2 ty), Tboolean
      else
	error loc (Other "invalid operands to comparison")
  | PEcall ({id_name=f}, el) ->
      begin try 
	let f = Hashtbl.find functions f in 
	let params = List.map (fun x -> (x.lv_mode, x.lv_type)) f.f_params in
	let el = type_args env loc params el in
	Ecall (f, el), f.f_type
      with Not_found -> try
	let c = Hashtbl.find constructors f in 
	let params = List.map (fun p -> Value, p) c.c_args in
	let el = type_args env loc params el in
	Econstr (c, List.map snd el), Tid c.c_type 
      with Not_found ->
	error loc (UnboundFun f)
      end

and type_args env loc params el =
  let rec check_args tel' = function
    | [], [] -> 
	List.rev tel'
    | { p_expr_loc = loc } as e :: el, (Reference, ty) :: tl ->
	begin match type_expr env e with
	  | {t_expr_node = Elvalue _} as te -> 
	      check_args ((Reference, coerce loc te ty) :: tel') (el, tl)
	  | _ -> error loc (Other "reference expected, value given")
	end
    | e :: el, (Value, ty) :: tl ->
	let te = type_expr env e in
	check_args ((Value, coerce e.p_expr_loc te ty) :: tel') (el, tl)
    | [], _ :: _ ->
	errors loc "too few arguments"
    | _, [] ->
	errors loc "too many arguments"
  in
  check_args [] (el, params)

and type_lvalue env loc = function
  | PLident x -> 
      begin
	try 
	  let l = Env.find x.id_name env in Llocal l, l.lv_type
	with Not_found -> try
	  let g = Hashtbl.find global_vars x.id_name in Lglobal g, g.gv_type
	with Not_found ->
	  error x.id_loc (UnboundVar x.id_name)
      end
  | PLarrget (lv, e) ->
      let lv, ty = type_lvalue env loc lv in
      let e = type_int_expr env e in
      match ty with
	| Tarray (n1, _, ty) -> Larrget (lv, e, (n1, ty)), ty
	| _ -> error loc (Other "subscripted object is not an array")

and type_int_expr env e =
  let te = type_expr env e in
  let ty = te.t_expr_type in
  if ty <> Tinteger then error e.p_expr_loc (ExpectedType (ty, Tinteger));
  te

and type_boolean_expr env e =
  let te = type_expr env e in
  let ty = te.t_expr_type in
  if ty <> Tboolean then error e.p_expr_loc (ExpectedType (ty, Tboolean));
  te

and type_num_expr env e =
  let te = type_expr env e in
  let ty = te.t_expr_type in
  if ty <> Tinteger && ty <> Treal then error e.p_expr_loc (ExpectedNum ty);
  te

(* Instructions *)

(* indique s'il s'agit de la fonction en cours de dfinition *)
let is_the_function x = function
  | None -> false
  | Some (y, _) -> x = y

let rec type_statement env et st = 
  type_statement_node env et st.p_st_loc st.p_st_node

and type_statement_node env et loc = function
  | PSskip ->
      Sskip, false
  | PSassign (PLident {id_name=x}, e) when is_the_function x et ->
      let te = type_expr env e in
      begin match et with
	| Some (_, ty) -> Sassign_result (coerce e.p_expr_loc te ty), true
	| None -> assert false
      end
  | PSassign (lv, e) ->
      let lv, ty = type_lvalue env loc lv in
      if is_array_type ty then errors loc "array assignment is not allowed";
      let te = type_expr env e in
      Sassign (lv, coerce e.p_expr_loc te ty), false
  | PSprocedure_call ({id_name="writeln"}, el) ->
      let type_arg e = 
	let te = type_expr env e in
	match te.t_expr_type with
	  | Tboolean | Tinteger | Treal -> 
	      te
	  | _ -> 
	      error e.p_expr_loc (Other "argument of writeln is of wrong type")
      in
      let el = List.map type_arg el in
      Swriteln el, false
  | PSprocedure_call ({id_name = p}, el) ->
      if not (Hashtbl.mem procedures p) then error loc (UnboundProc p);
      let p = Hashtbl.find procedures p in
      let params = List.map (fun x -> (x.lv_mode, x.lv_type)) p.pr_params in
      let el = type_args env loc params el in
      Sprocedure_call (p, el), false
  | PSif (e, s1, s2) ->
      let e = type_boolean_expr env e in
      let s1,r1 = type_statement env et s1 in
      let s2,r2 = type_statement env et s2 in
      Sif (e, s1, s2), r1 && r2
  | PSwhile (e, s) ->
      let e = type_boolean_expr env e in
      let s,_ = type_statement env et s in
      Swhile (e, s), false
  | PSfor (i, e1, e2, s) ->
      let lv,ty = type_lvalue env i.id_loc (PLident i) in
      if ty <> Tinteger then error i.id_loc (ExpectedType (ty, Tinteger));
      let e1 = type_int_expr env e1 in
      let e2 = type_int_expr env e2 in
      let s,_ = type_statement env et s in
      let v2 = value_var "for_bound" Tinteger in
      Sfor (lv, v2, e1, e2, s), false
  | PSbloc sl ->
      let sl,r = type_block env et sl in
      Sbloc sl, r
  | PSmatch (e, bl) ->
      let e = type_expr env e in
      let ty = match e.t_expr_type with
	| Tarray _ -> errors loc "can't match over an array"
	| ty -> ty
      in
      let bl,r = type_branches env et ty bl in
      let c =
	try 
	  Matching.compile ty bl
	with 
	  | NotExhaustive -> errors loc "this matching is not exhaustive"
      in
      Smatch (e, c), r

and type_block env et = function
  | [] -> 
      [], false
  | s :: sl -> 
      let s,r = type_statement env et s in
      let sl,rl = type_block env et sl in
      s :: sl, r || rl

and type_branches env et ty = function
  | [] ->
      [], true
  | b :: bl -> 
      let b,r = type_branch env et ty b in
      let bl,rl = type_branches env et ty bl in
      b :: bl, r && rl

and type_branch env et ty (p, sl) =
  let p, env = type_pattern env et ty p in
  let sl,r = type_block env et sl in
  (p, sl), r

and type_pattern env et ty p =
  let loc = p.p_pattern_loc in
  match p.p_pattern_node with
  | PPall -> 
      Pall, env
  | PPconst c ->
      if ty <> type_constant c then error loc (ExpectedPattern ty);
      Pconstructor (Cconst c, []), env
  | PPid {id_name=x} when Hashtbl.mem constructors x ->
      let c = Hashtbl.find constructors x in
      if Tid c.c_type <> ty then error loc (ExpectedPattern ty);
      if List.length c.c_args > 0 then errors loc "too few arguments";
      Pconstructor (Cnamed c, []), env
  | PPid {id_name=x} ->
      begin match et with 
	| Some (f,_) when x = f -> error loc (Shadow x)
	| _ -> ()
      end;
      if Env.mem x env then error loc (Clash x);
      let v = 
	{ lv_name = x; lv_type = ty; lv_mode = Value; lv_id = unique_id () } 
      in
      Pid v, Env.add x v env
  | PPconstructor ({id_name=id; id_loc=idloc}, pl) ->
      if not (Hashtbl.mem constructors id) then error idloc (UnboundConstr id);
      let c = Hashtbl.find constructors id in
      if Tid c.c_type <> ty then error loc (ExpectedPattern ty);
      let nbc = List.length c.c_args in
      let nbp = List.length pl in
      if nbc < nbp then errors loc "too many arguments";
      if nbc > nbp then errors loc "too few arguments";
      let pl,env = 
	fold_map 
	  (fun (p,ty) env -> type_pattern env et ty p) 
	  (List.combine pl c.c_args) env
      in
      Pconstructor (Cnamed c, pl), env

(* Variables globales *)

let decl_global_var ({id_name=id; id_loc=loc}, ty) =
  if is_global id then error loc (Clash id);
  let gv = { gv_name = id; gv_type = type_type ty; gv_id = unique_id () } in
  Hashtbl.add global_vars id gv;
  gv

(* Fonctions et procdures *)

let definition (id, pl, d, vars, st) = 
  if is_global id.id_name then error id.id_loc (Clash id.id_name);
  let is_function = match d with PDfunction _ -> true | _ -> false in
  let local_var m ({id_name=x; id_loc=loc}, ty) env =
    if is_function && x = id.id_name then error loc (Shadow x);
    if Env.mem x env then error loc (Clash x);
    let v = 
      { lv_name = x; lv_type = type_type ty; 
	lv_mode = m; lv_id = unique_id () } 
    in
    v, Env.add x v env
  in
  let formal_param (m,x,t) env = local_var m (x,t) env in
  let pl,env = fold_map formal_param pl Env.empty in
  let vars, env =  fold_map (local_var Value) vars env in
  match d with
    | PDfunction ty -> 
	let ty = type_type ty in
	if is_array_type ty then 
	  errors id.id_loc "a function can't return an array";
	let f = { f_name = id.id_name; f_params = pl; f_type = ty } in
	Hashtbl.add functions id.id_name f;
	let st,r = type_statement env (Some (id.id_name, ty)) st in
	if not r then error id.id_loc NoReturn;
	Dfunction (f, { b_vars = vars; b_stmt = st })
    | PDprocedure ->
	let p = { pr_name = id.id_name; pr_params = pl } in
	Hashtbl.add procedures id.id_name p;
	let st,_ = type_statement env None st in
	Dprocedure (p, { b_vars = vars; b_stmt = st })

let file f = 
  let dt = List.map decl_type f.p_decl_types in
  let dv = List.map decl_global_var f.p_decl_vars in
  let dd = List.map definition f.p_definitions in
  let s,_ = type_statement Env.empty None f.p_main in
  {
    t_name = f.p_name.id_name;
    t_decl_types = dt;
    t_decl_vars = dv;
    t_definitions = dd;
    t_main = s;
  }
  
